package Lpatch;
#use Exporter();
#use Plist;
#@ISA = qw(Exporter);
use Cwd;
use IO::Handle;

my $Patlen = undef;
my $InstallDir = cwd();

sub new {
        my ($proto, $dirpath) = @_;
        my $self  = {};
	$self->{Patched} = [];
	$self->{Stack} = [];
	$self->{Filename} = undef;
	$self->{Text} = undef;
	$self->{Position} = undef;
	$self->{OldBlock} = undef;
	$self->{NewBlock} = undef;
	$self->{StartDir} = cwd();
	$self->{SearchSkip} = 0;
	if ($dirpath) { chdir $dirpath or die "Couldn\'t chdir to $dirpath\n";}
        bless($self);   # same as return
}

sub DESTROY {
	my $self = shift;
	&Lpatch::close($self);
	chdir $self->{StartDir};
	if (@{$self->{Patched}} > 0) { print STDERR "Patch wasn\'t apply.\n"; }
}

sub do {
	my ($self, $prefix, $postfix) = @_;
	my ($filename, $text);
	&Lpatch::close($self);
	if (!defined($prefix)) { $prefix = ""; }
	if (!defined($postfix)) { $postfix = ""; }
	while (@{$self->{Patched}}) {
		$filename = shift @{$self->{Patched}};
		$text = shift @{$self->{Patched}};
		open(FD, ">$prefix$filename$postfix") or
			die "Can\'t open $prefix$filename$postfix\n";
		foreach (@$text) { print FD $_; }
		close FD;
	}
}

sub open {
	my ($self, $filename, $pat) = @_;
	&Lpatch::close($self);
	$self->{Filename} = $filename;
	open(FD, "+<$filename") or die "Can\'t open $filename.\n";
	$self->{Text} = [<FD>];
	close FD;
	$self->{Position} = 0;  $self->{SearchSkip} = 0;
	$pat && &Lpatch::search($self, $pat, 1) &&
		die "$filename already patched.\n";
	$self->{Position} = 0;  $self->{SearchSkip} = 0;
}

sub close {
	my $self = shift;
	if (@{$self->{Stack}} > 0) {
	   die "Not closed block in $self->{Filename}\n";
	}
	if ($self->{Filename}) {
	   push @{$self->{Patched}}, $self->{Filename}, $self->{Text};
	   $self->{Filename} = undef;
	}
}

sub search {
	my ($self, $patterns, $mode) = @_; # mode used by open
	if (!defined($patterns)) { # if not def - pos to the end of text
	   return ($self->{Position} = @{$self->{Text}} - 1);
	}
	if (!ref($patterns)) { $patterns = [ $patterns ]; }
#print 'search ', $patterns->[0], "\n";
	$Patlen = @$patterns;
	my ($i, $j);
	$self->{Position} += $self->{SearchSkip};
	$self->{SearchSkip} = 1;
	GL: for ($i = $self->{Position}; $i < @{$self->{Text}}; $i++) {
		for ($j = 0; $j < @$patterns; $j++) {
#print "==>", ${$self->{Text}}[$i+$j], "<==\n";
		 if (!(${$self->{Text}}[$i+$j] =~ /$$patterns[$j]/)){ next GL; }
		}
		# found !
		$self->{Position} = $i; return $i;
	}
	$mode || die "Can\'t find @$patterns in $self->{Filename}\n";
	return undef;
}

sub insert {
	my ($self, $where, $str) = @_;
#	my (@tmp, $blen, $str1);
	if (!($str =~ /\n$/)) { $str .= "\n"; }
#	$blen = @tmp = split(/\n/, $str, -1); $blen -= 1; # number of ins-ed str
	splice @{$self->{Text}}, $self->{Position} + $where, 0, $str;
# if $where > 1 set Position after inserted strings!!!
	if ($where > 1) { $self->{Position} += $where;
	} else { $self->{Position} += 1; }
	$self->{SearchSkip} = 1;
}

sub insert_block {
	my ($self, $where, $block) = @_;
	if ($self->{OldBlock}) {
		$block = lc($block);
		if ($block eq "new") { $block = "NewBlock"; }
		elsif ($block eq "old") { $block = "OldBlock"; }
		else { die "Wrong block keyword in insert_block\n"; }
		&Lpatch::insert($self, $where, $self->{$block});
	} else { die "Block not defined in $self->{Filename}\n"; }
}

sub mkstr {
	my $str = join('',@{$_[0]});
	if (!($str =~ /\n$/)) { $str .= "\n"; }
	return $str;
}

sub block_end {
	my $self = shift;
	$self->{NewBlock} = &Lpatch::mkstr($self->{Text});
	$self->{OldBlock} = pop @{$self->{Stack}};
	$self->{Text} = pop @{$self->{Stack}};
	$self->{Position} = pop @{$self->{Stack}};
	$self->{SearchSkip} = 0;
}

sub block {
	# if $end == undef - block consists of lines, mapped by start pattern
	my ($self, $start, $end) = @_;
	my $endpos;
	my $startpos = &Lpatch::search($self, $start);
	if ($end) {
		$endpos = &Lpatch::search($self, $end) + $Patlen - 1;
	} else {
		$endpos = $startpos + $Patlen - 1;
	}
	push @{$self->{Stack}}, $startpos; # Position
	push @{$self->{Stack}}, # Text without block
	  [@{$self->{Text}}[0..$startpos-1,$endpos+1..@{$self->{Text}}-1]];
	$self->{Position} = 0;
	$self->{Text} = [@{$self->{Text}}[$startpos..$endpos]];
	push @{$self->{Stack}}, &Lpatch::mkstr($self->{Text}); #OldBlock
	$self->{SearchSkip} = 0;
}

sub setparam {
        my ($parms, $files) = @_;
        my @names = keys %$parms;
        my (@txt, $file, $name);
        foreach $file (@$files) {
		chomp($file);
		if ($file =~ /[ \t]/) { # it isn't filename, but err msg
			print "$file\n"; next;
		}
		if ( -d $file ) {
			my @a = <$file/*>;
			setparam($params, \@a);
			next;
		}
                @txt = ();
                CORE::open(FD, "+<$file") or die "Can't open $file for update.\n";
                while (<FD>) {
                  foreach $name (@names) {
                    s|$name|$$parms{$name}|g;
                  }
                  push(@txt, $_);
                }
                truncate(FD, 0);
                seek(FD,0,0);
                print FD @txt;
                CORE::close (FD);
        }
}

sub ask {
        my ($str, $default) = @_;
        my $answer = "";
        while ($answer eq "") {
           print $str, " [";
           if ($default) { print $default; }
           autoflush STDOUT 1;
           print "]: ";
           autoflush STDOUT 0;
           chomp($answer = <>);
           if (($answer eq "") && $default) { $answer = $default; }
        }
        return $answer;
}

sub get_arch_path {
# arch_desc - ref to list: ((arch_path, path for cpio -i, qwestion, param name)...)
	my $arch_desc = shift;
	my $dref;
	foreach $dref (@$arch_desc) {
	   -f $dref->[0] || die "No $dref->[0] archive\n";
	   do { $dref->[1] = ask($dref->[2], $dref->[1]); }
		until ( -d $dref->[1] ||
		   do { print "Invalid directory $dref->[1]\n"; 0; });
	}
}

sub get_params {
# param_desc - ref to list: ((param name, qwestion, condition, value)...)
	my $param_desc = shift;
	my ($pref, $param);
	foreach $pref (@$param_desc) {
	   $param = ask($pref->[1], $pref->[3]);
	   if ($param eq "") { redo; }
	   if (($pref->[2] eq "d") && !(-d $param)) {
		print "Invalid directory path $param\n"; redo; }
	   if (($pref->[2] eq "f") && !(-f $param)) {
		print "Invalid file path $param\n"; redo; }
	   if (($pref->[2] eq "x") && !(-x $param)) {
		print "Invalid executable path $param\n"; redo; }
	   $pref->[3] = $param;
	}
}

sub get_info {
	my ($archlist, $parmlist, $parms) = @_;
	get_arch_path($archlist);
	get_params($parmlist);
# all info gathered - build parms hash
	my $el;
	foreach $el (@$archlist) {
	   if (defined($el->[3])) {$parms->{$el->[3]} = $el->[1]; }}
	foreach $el (@$parmlist) { $parms->{$el->[0]} = $el->[3]; }
}

sub do_install {
	my ($archlist, $parms) = @_;
	my ($el, @files);
	foreach $el (@$archlist) {
	   chdir $el->[1];
	   @files = `cpio -idv -F $InstallDir/$el->[0] 2>&1`;
#	   `cpio -idv -F $InstallDir/$el->[0] 1>&2`; # stdout to stderr
	   setparam($parms, \@files);
	}
}

1;

=head1 NAME

Lpatch - yet another patch system

=head1 SYNOPSIS

use Lpatch;

$a = new Lpatch;
$a->open(filename [, pattern]);
$a->search(pattern);
$a->insert(pos, text); # or inline qwotation
$a->block(begin_pattern [, end_pattern]);
	# modify block with search and insert
$a->block_end;
$a->insert_block(pos, <"New" | "Old">);
$a->open(...);
...
$a->do(prefix, postfix);

&Lpatch::setparam(proto, files);

=head1 DESCRIPTION

open - opens file B<filename> for patch. To prevent double patch
of the same file B<pattern> may be defined. If B<pattern> match,
patch process terminates. B<pattern> is a Perl pattern string, or
reference to the list of such patterns. If it is a list - each
pattern in it will match the sequential lines of text.

search - positioning within text, set current position. If
position undfined - position to the end of text.

insert - insert B<text>, if pos == 0 - before, pos == 1 - after
current position.

block - remove block from text and modify it with search/insert
functions, until B<block_end>.

block_end - block removed, current position is on the first line
after the block.

insert_block - insert modified (New) or unmodified (Old) block.

do - upply modification to files. New filenames created as
B<prefix> + filename + B<postfix>. If there is no prefix, nor
postfix - modification inplace. All modifications prior B<do> are
made in memory, so if some modification couldn't be done, no
files will be changed.

setparam - insert prameters in files.

proto - reference to hash, keys are patterns which replaced by
value of this hash.
files - reference to array of filenames.

Serge Lozovsky, serge@intes.odessa.ua

=cut
